@@ -2,6 +2,6 @@ mod clip_buffer;
 mod editor;
 mod line_buffer;
 
-pub use clip_buffer::{get_default_clipboard, Clipboard};
+pub(crate) use clip_buffer::{get_default_clipboard, Clipboard, ClipboardMode};
 pub use editor::Editor;
 pub use line_buffer::LineBuffer;
