@@ -1,19 +1,16 @@
+use super::{base::HistoryNavigationQuery, History};
+use crate::core_editor::LineBuffer;
 use std::{
     collections::{vec_deque::Iter, VecDeque},
-    fs::{File, OpenOptions},
-    io::{BufRead, BufReader, BufWriter, Write},
+    fs::OpenOptions,
+    io::{BufRead, BufReader, BufWriter, Seek, SeekFrom, Write},
+    ops::{Deref, DerefMut},
     path::PathBuf,
 };
 
-use crate::core_editor::LineBuffer;
-
-use super::{
-    base::{HistoryAppender, HistoryNavigationQuery, HistoryView},
-    History,
-};
-
 /// Default size of the [`FileBackedHistory`] used when calling [`FileBackedHistory::default()`]
 pub const HISTORY_SIZE: usize = 1000;
+pub const NEWLINE_ESCAPE: &str = "<\\n>";
 
 /// Stateful history that allows up/down-arrow browsing with an internal cursor.
 ///
@@ -27,8 +24,7 @@ pub struct FileBackedHistory {
     entries: VecDeque<String>,
     cursor: usize, // If cursor == entries.len() outside history browsing
     file: Option<PathBuf>,
-    len_on_disk: usize,  // Keep track what was previously written to disk
-    truncate_file: bool, // as long as the file would not exceed capacity we can use appending writes
+    len_on_disk: usize, // Keep track what was previously written to disk
     query: HistoryNavigationQuery,
 }
 
@@ -41,18 +37,24 @@ impl Default for FileBackedHistory {
     }
 }
 
-impl History for FileBackedHistory {}
+fn encode_entry(s: &str) -> String {
+    s.replace('\n', NEWLINE_ESCAPE)
+}
 
-impl HistoryAppender for FileBackedHistory {
+fn decode_entry(s: &str) -> String {
+    s.replace(NEWLINE_ESCAPE, "\n")
+}
+
+impl History for FileBackedHistory {
     /// Appends an entry if non-empty and not repetition of the previous entry.
     /// Resets the browsing cursor to the default state in front of the most recent entry.
     ///
-    fn append(&mut self, entry: String) {
+    fn append(&mut self, entry: &str) {
         // Don't append if the preceding value is identical or the string empty
         if self
             .entries
             .back()
-            .map_or(true, |previous| previous != &entry)
+            .map_or(true, |previous| previous != entry)
             && !entry.is_empty()
         {
             if self.entries.len() == self.capacity {
@@ -60,19 +62,16 @@ impl HistoryAppender for FileBackedHistory {
                 // before adding a new one.
                 self.entries.pop_front();
                 self.len_on_disk = self.len_on_disk.saturating_sub(1);
-                self.truncate_file = true;
             }
-            self.entries.push_back(entry);
+            self.entries.push_back(entry.to_string());
         }
-        self.reset_cursor()
+        self.reset_cursor();
     }
 
     fn iter_chronologic(&self) -> Iter<'_, String> {
         self.entries.iter()
     }
-}
 
-impl HistoryView for FileBackedHistory {
     fn back(&mut self) {
         match self.query.clone() {
             HistoryNavigationQuery::Normal(_) => {
@@ -81,10 +80,10 @@ impl HistoryView for FileBackedHistory {
                 }
             }
             HistoryNavigationQuery::PrefixSearch(prefix) => {
-                self.back_with_criteria(&|entry| entry.starts_with(&prefix))
+                self.back_with_criteria(&|entry| entry.starts_with(&prefix));
             }
             HistoryNavigationQuery::SubstringSearch(substring) => {
-                self.back_with_criteria(&|entry| entry.contains(&substring))
+                self.back_with_criteria(&|entry| entry.contains(&substring));
             }
         }
     }
@@ -97,10 +96,10 @@ impl HistoryView for FileBackedHistory {
                 }
             }
             HistoryNavigationQuery::PrefixSearch(prefix) => {
-                self.forward_with_criteria(&|entry| entry.starts_with(&prefix))
+                self.forward_with_criteria(&|entry| entry.starts_with(&prefix));
             }
             HistoryNavigationQuery::SubstringSearch(substring) => {
-                self.forward_with_criteria(&|entry| entry.contains(&substring))
+                self.forward_with_criteria(&|entry| entry.contains(&substring));
             }
         }
     }
@@ -117,10 +116,26 @@ impl HistoryView for FileBackedHistory {
     fn get_navigation(&self) -> HistoryNavigationQuery {
         self.query.clone()
     }
+
+    fn query_entries(&self, search: &str) -> Vec<String> {
+        self.iter_chronologic()
+            .rev()
+            .filter(|entry| entry.contains(search))
+            .cloned()
+            .collect::<Vec<String>>()
+    }
+
+    fn max_values(&self) -> usize {
+        self.entries.len()
+    }
 }
 
 impl FileBackedHistory {
     /// Creates a new in-memory history that remembers `n <= capacity` elements
+    ///
+    /// # Panics
+    ///
+    /// If `capacity == usize::MAX`
     pub fn new(capacity: usize) -> Self {
         if capacity == usize::MAX {
             panic!("History capacity too large to be addressed safely");
@@ -131,7 +146,6 @@ impl FileBackedHistory {
             cursor: 0,
             file: None,
             len_on_disk: 0,
-            truncate_file: true,
             query: HistoryNavigationQuery::Normal(LineBuffer::default()),
         }
     }
@@ -150,50 +164,10 @@ impl FileBackedHistory {
             std::fs::create_dir_all(base_dir)?;
         }
         hist.file = Some(file);
-        hist.load_file()?;
+        hist.sync()?;
         Ok(hist)
     }
 
-    /// Loads history from the associated newline separated file
-    ///
-    /// Expects the [`History`] to be empty.
-    ///
-    ///
-    /// **Side effect:** creates not yet existing file.
-    fn load_file(&mut self) -> std::io::Result<()> {
-        let f = File::open(
-            self.file
-                .as_ref()
-                .expect("History::load_file should only be called if a filename is set"),
-        );
-        assert!(
-            self.entries.is_empty(),
-            "History currently designed to load file once in the constructor"
-        );
-        match f {
-            Err(e) => match e.kind() {
-                std::io::ErrorKind::NotFound => {
-                    File::create(self.file.as_ref().unwrap())?;
-                    Ok(())
-                }
-                _ => Err(e),
-            },
-            Ok(file) => {
-                let reader = BufReader::new(file);
-                let mut from_file: VecDeque<String> = reader.lines().map(Result::unwrap).collect();
-                let from_file = if from_file.len() > self.capacity {
-                    from_file.split_off(from_file.len() - self.capacity)
-                } else {
-                    from_file
-                };
-                self.len_on_disk = from_file.len();
-                self.entries = from_file;
-                self.reset_cursor();
-                Ok(())
-            }
-        }
-    }
-
     fn back_with_criteria(&mut self, criteria: &dyn Fn(&str) -> bool) {
         if !self.entries.is_empty() {
             let previous_match = self.entries.get(self.cursor);
@@ -206,7 +180,7 @@ impl FileBackedHistory {
                 .find(|(_, entry)| criteria(entry) && previous_match != Some(entry))
             {
                 // set to entry
-                self.cursor = next_cursor
+                self.cursor = next_cursor;
             }
         }
     }
@@ -221,41 +195,74 @@ impl FileBackedHistory {
             .find(|(_, entry)| criteria(entry) && previous_match != Some(entry))
         {
             // set to entry
-            self.cursor = next_cursor
+            self.cursor = next_cursor;
         } else {
-            self.reset_cursor()
+            self.reset_cursor();
         }
     }
 
     /// Writes unwritten history contents to disk.
     ///
     /// If file would exceed `capacity` truncates the oldest entries.
-    fn flush(&mut self) -> std::io::Result<()> {
-        if self.file.is_none() {
-            return Ok(());
+    pub fn sync(&mut self) -> std::io::Result<()> {
+        if let Some(fname) = &self.file {
+            // The unwritten entries
+            let own_entries = self.entries.range(self.len_on_disk..);
+
+            let mut f_lock = fd_lock::RwLock::new(
+                OpenOptions::new()
+                    .create(true)
+                    .write(true)
+                    .read(true)
+                    .open(fname)?,
+            );
+            let mut writer_guard = f_lock.write()?;
+            let (mut foreign_entries, truncate) = {
+                let reader = BufReader::new(writer_guard.deref());
+                let mut from_file = reader
+                    .lines()
+                    .map(|o| o.map(|i| decode_entry(&i)))
+                    .collect::<Result<VecDeque<_>, _>>()?;
+                if from_file.len() + own_entries.len() > self.capacity {
+                    (
+                        from_file.split_off(from_file.len() - (self.capacity - own_entries.len())),
+                        true,
+                    )
+                } else {
+                    (from_file, false)
+                }
+            };
+
+            {
+                let mut writer = BufWriter::new(writer_guard.deref_mut());
+                if truncate {
+                    writer.seek(SeekFrom::Start(0))?;
+
+                    for line in &foreign_entries {
+                        writer.write_all(encode_entry(line).as_bytes())?;
+                        writer.write_all("\n".as_bytes())?;
+                    }
+                } else {
+                    writer.seek(SeekFrom::End(0))?;
+                }
+                for line in own_entries {
+                    writer.write_all(encode_entry(line).as_bytes())?;
+                    writer.write_all("\n".as_bytes())?;
+                }
+                writer.flush()?;
+            }
+            if truncate {
+                let file = writer_guard.deref_mut();
+                let file_len = file.stream_position()?;
+                file.set_len(file_len)?;
+            }
+
+            let own_entries = self.entries.drain(self.len_on_disk..);
+            foreign_entries.extend(own_entries);
+            self.entries = foreign_entries;
+
+            self.len_on_disk = self.entries.len();
         }
-        let file = if self.truncate_file {
-            // Rewrite the whole file if we truncated the old output
-            self.len_on_disk = 0;
-            // TODO: make this file race safe if multiple instances are used.
-            OpenOptions::new()
-                .write(true)
-                .truncate(true)
-                .open(self.file.as_ref().unwrap())?
-        } else {
-            // If the file is not beyond capacity just append new stuff
-            // (use the stored self.len_on_disk as offset)
-            OpenOptions::new()
-                .append(true)
-                .open(self.file.as_ref().unwrap())?
-        };
-        let mut writer = BufWriter::new(file);
-        for line in self.entries.range(self.len_on_disk..) {
-            writer.write_all(line.as_bytes())?;
-            writer.write_all("\n".as_bytes())?;
-        }
-        writer.flush()?;
-        self.len_on_disk = self.entries.len();
 
         Ok(())
     }
@@ -269,14 +276,13 @@ impl FileBackedHistory {
 impl Drop for FileBackedHistory {
     /// On drop the content of the [`History`] will be written to the file if specified via [`FileBackedHistory::with_file()`].
     fn drop(&mut self) {
-        let _ = self.flush();
+        let _res = self.sync();
     }
 }
 
 #[cfg(test)]
 mod tests {
     use pretty_assertions::assert_eq;
-    use std::io::BufRead;
 
     use super::*;
 
@@ -303,8 +309,8 @@ mod tests {
     #[test]
     fn going_backwards_bottoms_out() {
         let mut hist = FileBackedHistory::default();
-        hist.append("command1".to_string());
-        hist.append("command2".to_string());
+        hist.append("command1");
+        hist.append("command2");
         hist.back();
         hist.back();
         hist.back();
@@ -316,8 +322,8 @@ mod tests {
     #[test]
     fn going_forwards_bottoms_out() {
         let mut hist = FileBackedHistory::default();
-        hist.append("command1".to_string());
-        hist.append("command2".to_string());
+        hist.append("command1");
+        hist.append("command2");
         hist.forward();
         hist.forward();
         hist.forward();
@@ -329,25 +335,25 @@ mod tests {
     #[test]
     fn appends_only_unique() {
         let mut hist = FileBackedHistory::default();
-        hist.append("unique_old".to_string());
-        hist.append("test".to_string());
-        hist.append("test".to_string());
-        hist.append("unique".to_string());
+        hist.append("unique_old");
+        hist.append("test");
+        hist.append("test");
+        hist.append("unique");
         assert_eq!(hist.entries.len(), 3);
     }
     #[test]
     fn appends_no_empties() {
         let mut hist = FileBackedHistory::default();
-        hist.append("".to_string());
+        hist.append("");
         assert_eq!(hist.entries.len(), 0);
     }
 
     #[test]
     fn prefix_search_works() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("find me as well"));
-        hist.append(String::from("test"));
-        hist.append(String::from("find me"));
+        hist.append("find me as well");
+        hist.append("test");
+        hist.append("find me");
 
         hist.set_navigation(HistoryNavigationQuery::PrefixSearch("find".to_string()));
 
@@ -360,9 +366,9 @@ mod tests {
     #[test]
     fn prefix_search_bottoms_out() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("find me as well"));
-        hist.append(String::from("test"));
-        hist.append(String::from("find me"));
+        hist.append("find me as well");
+        hist.append("test");
+        hist.append("find me");
 
         hist.set_navigation(HistoryNavigationQuery::PrefixSearch("find".to_string()));
         hist.back();
@@ -378,9 +384,9 @@ mod tests {
     #[test]
     fn prefix_search_returns_to_none() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("find me as well"));
-        hist.append(String::from("test"));
-        hist.append(String::from("find me"));
+        hist.append("find me as well");
+        hist.append("test");
+        hist.append("find me");
 
         hist.set_navigation(HistoryNavigationQuery::PrefixSearch("find".to_string()));
         hist.back();
@@ -398,10 +404,10 @@ mod tests {
     #[test]
     fn prefix_search_ignores_consecutive_equivalent_entries_going_backwards() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("find me as well"));
-        hist.append(String::from("find me once"));
-        hist.append(String::from("test"));
-        hist.append(String::from("find me once"));
+        hist.append("find me as well");
+        hist.append("find me once");
+        hist.append("test");
+        hist.append("find me once");
 
         hist.set_navigation(HistoryNavigationQuery::PrefixSearch("find".to_string()));
         hist.back();
@@ -413,10 +419,10 @@ mod tests {
     #[test]
     fn prefix_search_ignores_consecutive_equivalent_entries_going_forwards() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("find me once"));
-        hist.append(String::from("test"));
-        hist.append(String::from("find me once"));
-        hist.append(String::from("find me as well"));
+        hist.append("find me once");
+        hist.append("test");
+        hist.append("find me once");
+        hist.append("find me as well");
 
         hist.set_navigation(HistoryNavigationQuery::PrefixSearch("find".to_string()));
         hist.back();
@@ -431,11 +437,11 @@ mod tests {
     #[test]
     fn substring_search_works() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("substring"));
-        hist.append(String::from("don't find me either"));
-        hist.append(String::from("prefix substring"));
-        hist.append(String::from("don't find me"));
-        hist.append(String::from("prefix substring suffix"));
+        hist.append("substring");
+        hist.append("don't find me either");
+        hist.append("prefix substring");
+        hist.append("don't find me");
+        hist.append("prefix substring suffix");
 
         hist.set_navigation(HistoryNavigationQuery::SubstringSearch(
             "substring".to_string(),
@@ -457,7 +463,7 @@ mod tests {
     #[test]
     fn substring_search_with_empty_value_returns_none() {
         let mut hist = FileBackedHistory::default();
-        hist.append(String::from("substring"));
+        hist.append("substring");
 
         hist.set_navigation(HistoryNavigationQuery::SubstringSearch("".to_string()));
 
@@ -466,11 +472,10 @@ mod tests {
 
     #[test]
     fn writes_to_new_file() {
-        use std::fs::File;
-        use std::io::BufReader;
         use tempfile::tempdir;
 
         let tmp = tempdir().unwrap();
+        // check that it also works for a path where the directory has not been created yet
         let histfile = tmp.path().join("nested_path").join(".history");
 
         let entries = vec!["test", "text", "more test text"];
@@ -478,17 +483,241 @@ mod tests {
         {
             let mut hist = FileBackedHistory::with_file(5, histfile.clone()).unwrap();
 
-            entries.iter().for_each(|e| hist.append(e.to_string()));
+            entries.iter().for_each(|e| hist.append(e));
 
             // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
         }
 
-        let f = File::open(histfile).unwrap();
-
-        let actual: Vec<String> = BufReader::new(f).lines().map(|x| x.unwrap()).collect();
+        let reading_hist = FileBackedHistory::with_file(5, histfile).unwrap();
 
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
         assert_eq!(entries, actual);
 
         tmp.close().unwrap();
     }
+
+    #[test]
+    fn persists_newlines_in_entries() {
+        use tempfile::tempdir;
+
+        let tmp = tempdir().unwrap();
+        let histfile = tmp.path().join(".history");
+
+        let entries = vec![
+            "test",
+            "multiline\nentry\nunix",
+            "multiline\r\nentry\r\nwindows",
+            "more test text",
+        ];
+
+        {
+            let mut writing_hist = FileBackedHistory::with_file(5, histfile.clone()).unwrap();
+
+            entries.iter().for_each(|e| writing_hist.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        let reading_hist = FileBackedHistory::with_file(5, histfile).unwrap();
+
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
+        assert_eq!(entries, actual);
+
+        tmp.close().unwrap();
+    }
+
+    #[test]
+    fn truncates_file_to_capacity() {
+        use tempfile::tempdir;
+
+        let tmp = tempdir().unwrap();
+        let histfile = tmp.path().join(".history");
+
+        let capacity = 5;
+        let initial_entries = vec!["test 1", "test 2"];
+        let appending_entries = vec!["test 3", "test 4"];
+        let expected_appended_entries = vec!["test 1", "test 2", "test 3", "test 4"];
+        let truncating_entries = vec!["test 5", "test 6", "test 7", "test 8"];
+        let expected_truncated_entries = vec!["test 4", "test 5", "test 6", "test 7", "test 8"];
+
+        {
+            let mut writing_hist =
+                FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            initial_entries.iter().for_each(|e| writing_hist.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        {
+            let mut appending_hist =
+                FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            appending_entries
+                .iter()
+                .for_each(|e| appending_hist.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+            let actual: Vec<_> = appending_hist.iter_chronologic().collect();
+            assert_eq!(expected_appended_entries, actual);
+        }
+
+        {
+            let mut truncating_hist =
+                FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            truncating_entries
+                .iter()
+                .for_each(|e| truncating_hist.append(e));
+
+            let actual: Vec<_> = truncating_hist.iter_chronologic().collect();
+            assert_eq!(expected_truncated_entries, actual);
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        let reading_hist = FileBackedHistory::with_file(capacity, histfile).unwrap();
+
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
+        assert_eq!(expected_truncated_entries, actual);
+
+        tmp.close().unwrap();
+    }
+
+    #[test]
+    fn truncates_too_large_file() {
+        use tempfile::tempdir;
+
+        let tmp = tempdir().unwrap();
+        let histfile = tmp.path().join(".history");
+
+        let overly_large_previous_entries = vec![
+            "test 1", "test 2", "test 3", "test 4", "test 5", "test 6", "test 7", "test 8",
+        ];
+        let expected_truncated_entries = vec!["test 4", "test 5", "test 6", "test 7", "test 8"];
+
+        {
+            let mut writing_hist = FileBackedHistory::with_file(10, histfile.clone()).unwrap();
+
+            overly_large_previous_entries
+                .iter()
+                .for_each(|e| writing_hist.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        {
+            let truncating_hist = FileBackedHistory::with_file(5, histfile.clone()).unwrap();
+
+            let actual: Vec<_> = truncating_hist.iter_chronologic().collect();
+            assert_eq!(expected_truncated_entries, actual);
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        let reading_hist = FileBackedHistory::with_file(5, histfile).unwrap();
+
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
+        assert_eq!(expected_truncated_entries, actual);
+
+        tmp.close().unwrap();
+    }
+
+    #[test]
+    fn concurrent_histories_dont_erase_eachother() {
+        use tempfile::tempdir;
+
+        let tmp = tempdir().unwrap();
+        let histfile = tmp.path().join(".history");
+
+        let capacity = 7;
+        let initial_entries = vec!["test 1", "test 2", "test 3", "test 4", "test 5"];
+        let entries_a = vec!["A1", "A2", "A3"];
+        let entries_b = vec!["B1", "B2", "B3"];
+        let expected_entries = vec!["test 5", "B1", "B2", "B3", "A1", "A2", "A3"];
+
+        {
+            let mut writing_hist =
+                FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            initial_entries.iter().for_each(|e| writing_hist.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        {
+            let mut hist_a = FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            {
+                let mut hist_b = FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+                entries_b.iter().for_each(|e| hist_b.append(e));
+
+                // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+            }
+            entries_a.iter().for_each(|e| hist_a.append(e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        let reading_hist = FileBackedHistory::with_file(capacity, histfile).unwrap();
+
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
+        assert_eq!(expected_entries, actual);
+
+        tmp.close().unwrap();
+    }
+
+    #[test]
+    fn concurrent_histories_are_threadsafe() {
+        use tempfile::tempdir;
+
+        let tmp = tempdir().unwrap();
+        let histfile = tmp.path().join(".history");
+
+        let num_threads = 16;
+        let capacity = 2 * num_threads + 1;
+
+        let initial_entries = (0..capacity).map(|i| format!("initial {i}"));
+
+        {
+            let mut writing_hist =
+                FileBackedHistory::with_file(capacity, histfile.clone()).unwrap();
+
+            initial_entries.for_each(|e| writing_hist.append(&e));
+
+            // As `hist` goes out of scope and get's dropped, its contents are flushed to disk
+        }
+
+        let threads = (0..num_threads)
+            .map(|i| {
+                let cap = capacity;
+                let hfile = histfile.clone();
+                std::thread::spawn(move || {
+                    let mut hist = FileBackedHistory::with_file(cap, hfile).unwrap();
+                    hist.append(&format!("A{}", i));
+                    hist.sync().unwrap();
+                    hist.append(&format!("B{}", i));
+                })
+            })
+            .collect::<Vec<_>>();
+
+        for t in threads {
+            t.join().unwrap();
+        }
+
+        let reading_hist = FileBackedHistory::with_file(capacity, histfile).unwrap();
+
+        let actual: Vec<_> = reading_hist.iter_chronologic().collect();
+
+        assert!(
+            actual.contains(&&format!("initial {}", capacity - 1)),
+            "Overwrote entry from before threading test"
+        );
+
+        for i in 0..num_threads {
+            assert!(actual.contains(&&format!("A{}", i)),);
+            assert!(actual.contains(&&format!("B{}", i)),);
+        }
+
+        tmp.close().unwrap();
+    }
 }
